# encoding: utf-8
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE.txt file distributed with
# this work for additional information regarding copyright ownership.  The ASF
# licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.    

'''Agile Object Oriented Data Technology (OODT).

OODT is a set of software components that permit the discovery, correlation,
search, and retrieval of data and metadata.  For more information, see the
OODT web site_.

.. _site: http://incubator.apache.org/projects/oodt.html
'''

__docformat__ = 'restructuredtext'



# encoding: utf-8
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE.txt file distributed with
# this work for additional information regarding copyright ownership.  The ASF
# licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.    

'''Unit tests for the `profile` module.
'''

__docformat__ = 'restructuredtext'

import unittest, xml.dom.minidom
from oodt.profile import ProfileAttributes, ResourceAttributes, Profile, UnspecifiedProfileElement, RangedProfileElement,\
    EnumeratedProfileElement

class ProfileAttributesTest(unittest.TestCase):
    '''Unit test for the ProfileAttributes class.
    '''
    def testDefaults(self):
        '''Test to see if default values are reasonable.
        '''
        pa = ProfileAttributes()
        self.assertEquals('UNKNOWN', pa.id)
        self.assertEquals('1.0.0', pa.version)
        self.assertEquals('profile', pa.type)
        self.assertEquals('active', pa.statusID)
        self.assertEquals('unclassified', pa.securityType)
        self.assertEquals('UNKNOWN', pa.parentID)
        self.assertEquals(0, len(pa.childIDs))
        self.assertEquals('UNKNOWN', pa.regAuthority)
        self.assertEquals(0, len(pa.revNotes))

    def testCmp(self):
        '''Test comparison operators.
        '''
        a = ProfileAttributes('1')
        b = ProfileAttributes('1')
        c = ProfileAttributes('2')
        self.assertEquals(a, a)
        self.assertEquals(a, b)
        self.assertNotEquals(a, c)
        self.assert_(a <= a)
        self.assert_(a <= b)
        self.assert_(a <= c)
        self.assert_(a < c)
        
    def testXML(self):
        '''Test XML serialization and re-composition from XML.
        '''
        a = ProfileAttributes('1.3.1.9', '2.0.0', 'profile', 'inactive', 'classified', '1.3.1', ['1.3.1.9.1', '1.3.1.9.2'],
            'NASA', ['Updated', 'Created'])
        doc = xml.dom.minidom.getDOMImplementation().createDocument(None, None, None)
        node = a.toXML(doc)
        b = ProfileAttributes(node=node)
        self.assertEquals(a, b)
        
    def testXMLValidity(self):
        '''Test to see if all required XML elements are in there.
        '''
        a = ProfileAttributes('1.3.1.9', '2.0.0', 'profile', 'inactive', 'classified', '1.3.1', ['1.3.1.9.1', '1.3.1.9.2'],
            'NASA', ['Updated', 'Created'])
        doc = xml.dom.minidom.getDOMImplementation().createDocument(None, None, None)
        node = a.toXML(doc)
        self.assertEquals('profAttributes', node.nodeName)
        childElements = [n.nodeName for n in node.childNodes]
        self.assertEquals([u'profId', u'profVersion', u'profType', u'profStatusId', u'profSecurityType', u'profParentId',
            u'profChildId', u'profChildId', u'profRegAuthority', u'profRevisionNote', u'profRevisionNote'],
            childElements)
    
    def testInstances(self):
        '''Test to ensure instances don't share instance data.
        '''
        a = ProfileAttributes(id='1')
        b = ProfileAttributes(id='2')
        self.assertNotEquals(a, b)
        a.childIDs.append('3')
        self.assertNotEquals(a.childIDs, b.childIDs)
        a.revNotes.append('Uhhhhh, spam?')
        self.assertNotEquals(a.revNotes, b.revNotes)
    
    
class ResourceAttributesTest(unittest.TestCase):
    '''Unit test for the ResourceAttributes class.
    '''
    def testDefaults(self):
        '''Test if default values are reasonable.
        '''
        ra = ResourceAttributes()
        self.assertEquals('UNKNOWN', ra.identifier)
        self.assertEquals('UNKNOWN', ra.title)
        self.assertEquals(0, len(ra.formats))
        self.assertEquals('UNKNOWN', ra.description)
        self.assertEquals(0, len(ra.creators))
        self.assertEquals(0, len(ra.subjects))
        self.assertEquals(0, len(ra.publishers))
        self.assertEquals(0, len(ra.contributors))
        self.assertEquals(0, len(ra.dates))
        self.assertEquals(0, len(ra.types))
        self.assertEquals(0, len(ra.sources))
        self.assertEquals(0, len(ra.languages))
        self.assertEquals(0, len(ra.relations))
        self.assertEquals(0, len(ra.coverages))
        self.assertEquals(0, len(ra.rights))
        self.assertEquals(0, len(ra.contexts))
        self.assertEquals('UNKNOWN', ra.aggregation)
        self.assertEquals('UNKNOWN', ra.resClass)
        self.assertEquals(0, len(ra.locations))
    
    def testCmp(self):
        '''Test comparison operations.
        '''
        a = ResourceAttributes('uri:fish', 'Fish', ['text/html'], 'A book about fish.')
        b = ResourceAttributes('uri:fish', 'Fish', ['text/html'], 'A book about fish.')
        c = ResourceAttributes('uri:clams', 'Clams', ['text/html'], 'A book about clams.')
        self.assertEquals(a, a)
        self.assertEquals(a, b)
        self.assertNotEquals(a, c)
        self.assert_(a <= b)
        self.assert_(a >= c)
        self.assert_(a > c)
        self.assert_(a <= a)
    
    def testXML(self):
        '''Test XML serialization and recomposition from XML.
        '''
        a = ResourceAttributes('uri:fish', 'Fish', ['text/html'], 'A book about fish.')
        doc = xml.dom.minidom.getDOMImplementation().createDocument(None, None, None)
        node = a.toXML(doc)
        b = ResourceAttributes(node=node)
        self.assertEquals(a, b)
    
    def testXMLValidity(self):
        '''Test to see if all required XML elements are in there.
        '''
        a = ResourceAttributes('uri:anus', 'Anus', ['text/html'], 'The anus, rectum, and other parts of the bum.',
            ['Buttman'], ['butts', 'henies', 'booties'], ['Butts and Co Publishing'], ['Dr Eugene Bottomman, III'],
            [], ['reference'], ['The Big Book of Booty'], ['en'], ['Buttholes and other oddities'],
            ['anatomy'], ['Cannot touch this'], ['system.buttServer'], 'granule', 'system.buttServer',
            ['http://butt.info/butt'])
        doc = xml.dom.minidom.getDOMImplementation().createDocument(None, None, None)
        node = a.toXML(doc)
        childElements = [n.nodeName for n in node.childNodes]       
        self.assertEquals([u'Identifier', u'Title', u'Format', u'Description', u'Creator', u'Subject', u'Subject',
            u'Subject', u'Publisher', u'Contributor', u'Type', u'Source', u'Language', u'Relation', u'Coverage',
            u'Rights', u'resContext', u'resAggregation', u'resClass', u'resLocation'], childElements)
    
    def testInstances(self):
        '''Test to ensure instances don't share instance data.
        '''
        a = ResourceAttributes()
        b = ResourceAttributes()
        a.formats.append('text/xml')
        a.creators.append('Dennis Moore')
        a.subjects.append('Silliness')
        a.publishers.append('BBC')
        a.contributors.append('Nigel')
        a.types.append('Video')
        a.languages.append('en')
        a.relations.append('Fawlty Towers')
        a.coverages.append('1970')
        a.rights.append('Abused')
        a.contexts.append('humor')
        a.locations.append('http://bbc.co.uk/')
        self.assertNotEquals(a.formats, b.formats)
        self.assertNotEquals(a.creators, b.creators)
        self.assertNotEquals(a.subjects, b.subjects)
        self.assertNotEquals(a.publishers, b.publishers)
        self.assertNotEquals(a.contributors, b.contributors)
        self.assertNotEquals(a.types, b.types)
        self.assertNotEquals(a.languages, b.languages)
        self.assertNotEquals(a.relations, b.relations)
        self.assertNotEquals(a.coverages, b.coverages)
        self.assertNotEquals(a.rights, b.rights)
        self.assertNotEquals(a.contexts, b.contexts)
        self.assertNotEquals(a.locations, b.locations)
    

class ProfileTest(unittest.TestCase):
    '''Unit test for class Profile.
    '''
    def testCmp(self):
        a = Profile()
        b = Profile()
        self.assertEquals(a, b)
        
    def testXML(self):
        '''Test XML serialization and recomposition from XML.
        '''
        a = Profile()
        doc = xml.dom.minidom.getDOMImplementation().createDocument(None, None, None)
        node = a.toXML(doc)
        b = Profile(node=node)
        self.assertEquals(a, b)

        x = UnspecifiedProfileElement('tastiness', 'How tasty it was', 'char', 'subjective', ['yumminess'],
            'This is highly subjective.')
        y = EnumeratedProfileElement('meal', 'What meal was eaten', 'char', 'meal', ['serving'], 'Typical values',
            ['Breakfast', 'Brunch', 'Lunch', 'Dinner'])
        z = RangedProfileElement('spicyness', 'How spicy it was', 'float', 'scovilles', ['piquancy'],
            'Hotter the better, I say', 0.0, 1000000.0)
        a.profElements['tastiness'], a.profElements['meal'], a.profElements['spicyness'] = x, y, z
        node = a.toXML(doc)
        b = Profile(node=node)
        self.assertEquals(a, b)

    def testInstances(self):
        '''Test to ensure isntances don't share isntance data.
        '''
        a = Profile()
        b = Profile()
        self.assertEquals(a, b)
        a.profElements['a'] = 'b'
        self.assertNotEquals(a, b)
    

def test_suite():
    '''Make the test suite.
    '''
    import doctest
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(ProfileAttributesTest))
    suite.addTest(unittest.makeSuite(ResourceAttributesTest))
    suite.addTest(unittest.makeSuite(ProfileTest))
    return suite
    

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# encoding: utf-8
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE.txt file distributed with
# this work for additional information regarding copyright ownership.  The ASF
# licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.    

'''Unit tests for the query module.

This module tests various query classes and the query expression parser.
'''

__docformat__ = 'restructuredtext'

import unittest, xml.dom, oodt.xmlutils
from oodt.query import QueryElement, QueryHeader, QueryResult, Query, _parseQuery, ExpressionParseError

class QueryElementTest(unittest.TestCase):
    '''Test the `QueryElement` class.
    '''
    def testDefaults(self):
        '''Test to see if defaults are reasonable.
        '''
        qe = QueryElement()
        self.assertEqual('UNKNOWN', qe.role)
        self.assertEqual('UNKNOWN', qe.value)

    def testArgs(self):
        '''Test to see if initializer arguments are used.
        '''
        qe = QueryElement(role='role', value='value')
        self.assertEqual('role', qe.role)
        self.assertEqual('value', qe.value)
    
    def testComparisons(self):
        '''Test comparison operators.
        '''
        a = QueryElement('a', '1')
        b = QueryElement('a', '1')
        c = QueryElement('b', '1')
        d = QueryElement('a', '2')
        self.assertEqual(a, a)
        self.assertEqual(a, b)
        self.assertEqual(b, a)
        self.assertNotEqual(a, c)
        self.assertNotEqual(c, a)
        self.assertNotEqual(a, d)
        self.assertNotEqual(d, a)
        self.assertNotEqual(c, d)
        self.assert_(a <= a)
        self.assert_(a <= c)
        self.assert_(a < c)
        self.assert_(a <= d)
        self.assert_(a < d)
    
    def testBadArgs(self):
        '''Test reactions to bad arugments.
        '''
        domImpl = xml.dom.getDOMImplementation()
        doc = domImpl.createDocument(None, None, None) # namespace URI, qualified Name, doctype
        self.assertRaises(ValueError, QueryElement, node=doc.createElement('notAQueryElement'))
    
    def testXML(self):
        '''Test XML serialization.
        '''
        domImpl = xml.dom.getDOMImplementation()
        doc = domImpl.createDocument(None, None, None) # namespace URI, qualified Name, doctype
        q1 = QueryElement(role='a', value='1')
        root = q1.toXML(doc)
        self.assertEqual('queryElement', root.nodeName)
        for child in root.childNodes:
            if 'tokenRole' == child.nodeName:
                self.assertEquals('a', oodt.xmlutils.text(child))
            elif 'tokenValue' == child.nodeName:
                self.assertEquals('1', oodt.xmlutils.text(child))
            else:
                self.fail('Unknown node "' + child.nodeName + '" in XML result')

        q2 = QueryElement(node=root)
        self.assertEqual(q1, q2)

        root = doc.createElement('queryElement')
        doc.appendChild(root)
        elem = doc.createElement('tokenRole')
        root.appendChild(elem)
        elem.appendChild(doc.createTextNode('a'))
        elem = doc.createElement('tokenValue')
        root.appendChild(elem)
        elem.appendChild(doc.createTextNode('2'))
        q3 = QueryElement(node=root)
        self.assertNotEqual(q1, q3)
    

_QUERY_HEADER_ATTRS = {
    'queryId': 'id',
    'queryTitle': 'title',
    'queryDesc': 'desc',
    'queryType': 'type',
    'queryStatusId': 'status',
    'querySecurityType': 'security',
    'queryRevisionNote': 'rev',
    'queryDataDictId': 'dataDict'
}

class QueryHeaderTest(unittest.TestCase):
    '''Unit test for the `QueryHeader` class.
    '''
    def testDefaults(self):
        '''Test if defaults are reasonable.
        '''
        qh = QueryHeader()
        self.assertEqual('UNKNOWN', qh.id)
        self.assertEqual('UNKNOWN', qh.title)
        self.assertEqual('UNKNOWN', qh.desc)
        self.assertEqual('QUERY', qh.type)
        self.assertEqual('ACTIVE', qh.status)
        self.assertEqual('UNKNOWN', qh.security)
        self.assertEqual('2005-10-01 SCK v0.0.0 Under Development', qh.rev)
        self.assertEqual('UNKNOWN', qh.dataDict)

    def testArgs(self):
        '''Test if initializer arguments are used.
        '''
        qh = QueryHeader('id', 'title', 'desc', 'type', 'status', 'security', 'rev', 'dataDict')
        self.assertEqual(qh.id, 'id')
        self.assertEqual(qh.title, 'title')
        self.assertEqual(qh.desc, 'desc')
        self.assertEqual(qh.type, 'type')
        self.assertEqual(qh.status, 'status')
        self.assertEqual(qh.security, 'security')
        self.assertEqual(qh.rev, 'rev')
        self.assertEqual(qh.dataDict, 'dataDict')

    def testBadArgs(self):
        '''Test reaction to bad arguments.
        '''
        domImpl = xml.dom.getDOMImplementation()
        doc = domImpl.createDocument(None, None, None) # namespace URI, qualified Name, doctype
        self.assertRaises(ValueError, QueryHeader, node=doc.createElement('notAQueryHeader'))

    def testXML(self):
        '''Test XML serialization.
        '''
        domImpl = xml.dom.getDOMImplementation()
        doc = domImpl.createDocument(None, None, None) # namespace URI, qualified Name, doctype
        q1 = QueryHeader('id', 'title', 'desc', 'type', 'status', 'security', 'rev', 'dataDict')
        root = q1.toXML(doc)
        self.assertEqual('queryAttributes', root.nodeName)
        for child in root.childNodes:
            self.check(child.nodeName, oodt.xmlutils.text(child))
        q2 = QueryHeader(node=root)
        self.assertEqual(q1, q2)

    def check(self, name, value):
        '''Check if the given tag name is valid.
        '''
        if name in _QUERY_HEADER_ATTRS:
            self.assertEqual(value, _QUERY_HEADER_ATTRS[name])
        else:
            fail('Unknown element ' + name + ' in query header')
    

class QueryResultTest(unittest.TestCase):
    '''Unit test for the `QueryResults` class.
    '''
    def testDefaults(self):
        '''Test if defaults are reasonable.
        '''
        qr = QueryResult()
        self.assertEquals(0, len(qr.results))
        self.assertEquals(0, len(qr))
        self.assertRaises(IndexError, qr.__getitem__, 0)

    def testBadArgs(self):
        '''Test reaction to bad arguments.
        '''
        domImpl = xml.dom.getDOMImplementation()
        doc = domImpl.createDocument(None, None, None) # namespace URI, qualified Name, doctype
        self.assertRaises(ValueError, QueryResult, node=doc.createElement('notAQueryResult'))

    def testXML(self):
        '''Test XML serialization.
        '''
        domImpl = xml.dom.getDOMImplementation()
        doc = domImpl.createDocument(None, None, None) # namespace URI, qualified Name, doctype
        q1 = QueryResult()
        root = q1.toXML(doc)
        self.assertEqual('queryResultSet', root.nodeName)
        q2 = QueryResult(node=root)
        self.assertEqual(q1, q2)
    

class QueryTest(unittest.TestCase):
    '''Unit test for the `Query` class.
    '''
    def testDefaults(self):
        '''Test if defaults are reasonable.
        '''
        q = Query()
        self.assertEqual(QueryHeader(), q.header)
        self.assertEqual('ATTRIBUTE', q.resultModeID)
        self.assertEqual('BROADCAST', q.propType)
        self.assertEqual('N/A', q.propLevels)
        self.assertEqual(1, q.maxResults)
        self.assertEqual(0, len(q.mimeAccept))

    def testParser(self):
        '''Test the query expresion parser.
        '''
        # Empty
        self.assertEqual(([], []), _parseQuery(''))
        
        # Simple
        self.assertEqual(([ QueryElement(role='elemName', value='x'), QueryElement(role='LITERAL', value='1'),
            QueryElement(role='RELOP', value='EQ') ], []), _parseQuery('x = 1'))

        # Logical or
        self.assertEqual(([ QueryElement(role='elemName', value='x'), QueryElement(role='LITERAL', value='1'),      
            QueryElement(role='RELOP', value='LT'), QueryElement(role='elemName', value='y'),
            QueryElement(role='LITERAL', value='-2'), QueryElement(role='RELOP', value='GT'),
            QueryElement(role='LOGOP', value='OR') ], []), _parseQuery('x < 1 or y > -2'))

        # Logical and has higher precendence
        self.assertEqual(([ QueryElement(role='elemName', value='x'), QueryElement(role='LITERAL', value='1'),
            QueryElement(role='RELOP', value='LE'), QueryElement(role='elemName', value='y'),
            QueryElement(role='LITERAL', value='2'), QueryElement(role='RELOP', value='GE'),
            QueryElement(role='elemName', value='z'), QueryElement(role='LITERAL', value='-3.96'),
            QueryElement(role='RELOP', value='NE'), QueryElement(role='LOGOP', value='AND'),
            QueryElement(role='LOGOP', value='OR') ], []), _parseQuery('x <= 1 | y >= 2 and z != -3.96'))
        
        # Logical or has lower precedence
        self.assertEqual(([ QueryElement(role='elemName', value='x'), QueryElement(role='LITERAL', value='1'),
            QueryElement(role='RELOP', value='LT'), QueryElement(role='elemName', value='y'),
            QueryElement(role='LITERAL', value='2'), QueryElement(role='RELOP', value='GT'),
            QueryElement(role='LOGOP', value='AND'), QueryElement(role='elemName', value='z'),
            QueryElement(role='LITERAL', value='3'), QueryElement(role='RELOP', value='NE'), 
            QueryElement(role='LOGOP', value='OR') ], []), _parseQuery('x LT 1 & y GT 2 or z NE 3'))

        # Parenthesis
        self.assertEqual(([ QueryElement(role='elemName', value='x'), QueryElement(role='LITERAL', value='1'),
            QueryElement(role='RELOP', value='LE'), QueryElement(role='elemName', value='y'), \
            QueryElement(role='LITERAL', value='2'), QueryElement(role='RELOP', value='GE'),
            QueryElement(role='LOGOP', value='OR'), QueryElement(role='elemName', value='z'),
            QueryElement(role='LITERAL', value='3'), QueryElement(role='RELOP', value='EQ'), 
            QueryElement(role='LOGOP', value='AND') ], []), _parseQuery('(x LE 1 or y GE 2) and z EQ 3'))

        # Logical not
        self.assertEqual(([ QueryElement(role='elemName', value='x'), QueryElement(role='LITERAL', value='1'),
            QueryElement(role='RELOP', value='LE'), QueryElement(role='elemName', value='y'),
            QueryElement(role='LITERAL', value='2'), QueryElement(role='RELOP', value='GE'),
            QueryElement(role='LOGOP', value='OR'), QueryElement(role='LOGOP', value='NOT'),
            QueryElement(role='elemName', value='z'), QueryElement(role='LITERAL', value='3'),
            QueryElement(role='RELOP', value='EQ'), QueryElement(role='LOGOP', value='NOT'),
            QueryElement(role='LOGOP', value='AND') ], []), _parseQuery('not (x LE 1 or y GE 2) and ! z EQ 3'))

        # Quoted strings
        self.assertEqual(([ QueryElement(role='elemName', value='x'), QueryElement(role='LITERAL', value='Fish Poop'),
            QueryElement(role='RELOP', value='EQ'), QueryElement(role='elemName', value='y'),
            QueryElement(role='LITERAL', value='Monkey Poop'), QueryElement(role='RELOP', value='EQ'),
            QueryElement(role='LOGOP', value='AND') ], []),
            _parseQuery("""x = "Fish Poop" and y = 'Monkey Poop'"""))
                
        # RETURN elements
        self.assertEqual(([ QueryElement(role='elemName', value='x'), QueryElement(role='LITERAL', value='Fish Poop'),
            QueryElement(role='RELOP', value='EQ'), QueryElement(role='elemName', value='y'),
            QueryElement(role='LITERAL', value='Monkey Poop'), QueryElement(role='RELOP', value='EQ'),
            QueryElement(role='LOGOP', value='AND') ], [QueryElement(role='elemName', value='fish')]),
            _parseQuery("""x = "Fish Poop" and RETURN > fish and y = 'Monkey Poop'"""))
        
        # RETURN at beginning and middle
        self.assertEqual(([ QueryElement(role='elemName', value='x'), QueryElement(role='LITERAL', value='Fish Poop'),
            QueryElement(role='RELOP', value='EQ'), QueryElement(role='elemName', value='y'),
            QueryElement(role='LITERAL', value='Monkey Poop'), QueryElement(role='RELOP', value='EQ'),
            QueryElement(role='LOGOP', value='AND') ], [QueryElement(role='elemName', value='fish'),
            QueryElement(role='elemName', value='poop')]),
            _parseQuery("""RETURN < fish and x = "Fish Poop" and RETURN > poop and y = 'Monkey Poop'"""))
            
        # Bad expressions
        for expr in ('heya', 'RETURN =', 'AND', 'LAT ~ fish', 'NOT', '(', ')'):
            self.assertRaises(ExpressionParseError, _parseQuery, expr)
            
        # Unusual symbols
        self.assertEqual(([QueryElement('elemName', 'DATA_SET_ID'), QueryElement('LITERAL', 'ARCB-L-RTLS-3-70CM-V1.0'),
            QueryElement('RELOP', 'EQ')], []), _parseQuery('''DATA_SET_ID = ARCB-L-RTLS-3-70CM-V1.0'''))
        self.assertEqual(([QueryElement('elemName', 'file'), QueryElement('LITERAL', '/usr/local/bin/poop'),
            QueryElement('RELOP', 'EQ')], []), _parseQuery('''file = /usr/local/bin/poop'''))
        self.assertEqual(([QueryElement('elemName', 'start_time'), QueryElement('LITERAL', '2006-02-06T13:12:13'),
            QueryElement('RELOP', 'EQ')], []), _parseQuery('''start_time = 2006-02-06T13:12:13'''))
    
    def testCmp(self):
        '''Test comparison operators.
        '''
        a = Query('lat > 3 and lon < -92.6')
        b = Query('lat > 3 and lon < -92.6')
        c = Query('lat < 3 and lon > -92.6')
        self.assert_(a == a)
        self.assert_(a == b)
        self.assert_(a < c)
        self.assert_(c > a)
        self.assert_(a <= b)
        self.assert_(a >= b)
        self.assert_(a <= c)
        self.assert_(c >= a)
        self.assert_(c != a)

    def testXML(self):
        '''Test XML serialization.
        '''
        domImpl = xml.dom.getDOMImplementation()
        doc = domImpl.createDocument(None, None, None) # namespace URI, qualified Name, doctype
        q1 = Query('lat > 3 and lon < -92.6')
        root = q1.toXML(doc)
        self.assertEqual('query', root.nodeName)
        q2 = Query(node=root)
        self.assertEqual(q1, q2)
    

def test_suite():
    '''Create the suite of tests.
    '''
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(QueryElementTest))
    suite.addTest(unittest.makeSuite(QueryHeaderTest))
    suite.addTest(unittest.makeSuite(QueryResultTest))
    suite.addTest(unittest.makeSuite(QueryTest))
    return suite
    

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
    



# encoding: utf-8
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE.txt file distributed with
# this work for additional information regarding copyright ownership.  The ASF
# licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.    

'''Agile OODT Tests.
'''

__docformat__ = 'restructuredtext'

import unittest
import profileTest, queryTest, xmlutilsTest

def test_suite():
    '''Create the suite of tests.
    '''
    suite = unittest.TestSuite()
    suite.addTest(profileTest.test_suite())
    suite.addTest(queryTest.test_suite())
    suite.addTest(xmlutilsTest.test_suite())
    return suite
    

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
    



# encoding: utf-8
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE.txt file distributed with
# this work for additional information regarding copyright ownership.  The ASF
# licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.    

'''Agile Web Grid.  HTTP-based profile and product servers.
'''

__docformat__ = 'restructuredtext'

from oodt.query import Query
from xml.dom.minidom import parseString, getDOMImplementation
from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
import cgi, os, shutil, stat

_validKinds = ('profile', 'product')
_doc = getDOMImplementation().createDocument(None, None, None) # qname, nsuri, doctype
    
class WebGridRequestHandler(BaseHTTPRequestHandler):
    '''HTTP request handler for Web-Grid requests.  This request handler accepts GET
    or POST requests and directs them to profile or product handlers.  Additionally,
    requests to install new handlers, to list currently installed handlers, and to
    remove handlers by ID are supported.
    '''
    def do_POST(self):
        '''Handle a POST.
        '''
        try:
            length = int(self.headers['content-length'])
            kind = self.headers['content-type']
            if kind.endswith('www-form-urlencoded'):
                if length > 0:
                    params = cgi.parse_qs(self.rfile.read(length), True, True) # Keep blanks, strict parse
                else:
                    params = {}
                self.__execute(self.path, params)
            else:
                raise ValueError('Unknown encoding "%s"' % kind)
        except Exception, e:
            self.send_error(500, str(e))
    
    def do_GET(self):
        '''Handle a GET.
        '''
        try:
            index = self.path.find('?')
            if index >= 0:
                params = cgi.parse_qs(self.path[index+1:], True, True) # Keep blanks, strict parse
                path = self.path[0:index]
            else:
                params, path = {}, self.path
            self.__execute(path, params)
        except Exception, e:
            self.send_error(500, str(e))
    
    def __execute(self, path, params):
        '''Execute an HTTP request.
        '''
        components = path.split('/')
        if len(components) == 3:
            context, command = components[1], components[2]
            if context != self.server.serverID:
                raise ValueError('Unknown server ID "%s"' % context)
            func = getattr(self, command)
            if callable(func):
                func(params)
                return
        raise KeyError('Unknown command')
    
    def echo(self, params):
        '''Debugging method that echoes back the request parameters.
        '''
        u = unicode(params)
        self.send_response(200)
        self.send_header('Content-type', 'text/plain;charset=utf-8')
        self.send_header('Content-length', str(len(u)))
        self.end_headers()
        self.wfile.write(u)
    
    def sendEmptyResponse(self):
        '''Send an empty response to the HTTP client.
        '''
        self.send_response(200)
        self.send_header('Content-type', 'text/plain;charset=utf-8')
        self.send_header('Content-length', '0')
        self.end_headers()
    
    def install(self, params):
        '''Install a new handler.  This will overwrite existing handlers with the
        same ID.
        '''
        handlers = self.server.getHandlers(params['kind'][0])
        globs = dict(globals())
        del globs['__name__']
        # TODO: use rexec or otherwise limit the code than can be uploaded.
        exec params['code'][0] in globs, globs
        handlers[params['id'][0]] = globs['handler']
        self.sendEmptyResponse()
    
    def remove(self, params):
        '''Remove an existing handler.
        '''
        handlers = self.server.getHandlers(params['kind'][0])
        del handlers[params['id'][0]]
        self.sendEmptyResponse()
    
    def list(self, params):
        '''List installed handlers.
        '''
        handlers = {}
        for kind in _validKinds:
            handlers[kind] = self.server.getHandlers(kind).keys()
        handlers = unicode(handlers)
        self.send_response(200)
        self.send_header('Content-type', 'text/plain;charset=utf-8')
        self.send_header('Content-length', str(len(handlers)))
        self.end_headers()
        self.wfile.write(handlers)
    
    def __createQuery(self, params):
        '''Create a Query from the request parameters.  This method prefers the
        xmlq parameter and parses it as an XML document and into a Query object.
        However, if it's not provided, or fails to parse, it'll use the q parameter,
        which is expected to be just a query expression.
        '''
        try:
            doc = parseString(params['xmlq'][0])
            return Query(node=doc.documentElement)
        except KeyError:
            return Query(params['q'][0])
    
    def sendProduct(self, match):
        '''Send a matching product.
        '''
        self.send_response(200)
        self.send_header('Content-type', match.contentType)
        self.send_header('Content-length', str(match.length))
        self.end_headers()
        shutil.copyfileobj(match.data, self.wfile)
        self.log_request(200, match.length)
    
    def prod(self, params):
        '''Handle a product query.
        '''
        query = self.__createQuery(params)
        for handler in self.server.getHandlers('product'):
            matches = handler.query(query)
            if len(matches) > 0:
                self.sendProduct(matches[0])
        self.send_error(404, 'No matching products')
    
    def prof(self, params):
        '''Handle a profile query.
        '''
        query = self.__createQuery(params)
        tmp = os.tmpfile()
        tmp.writelines((u'<?xml version="1.0" encoding="UTF-8"?>\n',
            u'<!DOCTYPE profiles PUBLIC "-//JPL//DTD Profile 1.1//EN"\n',
            u'  "http://oodt.jpl.nasa.gov/grid-profile/dtd/prof.dtd">\n',
            u'<profiles>\n'))
        for handler in self.server.getHandlers('profile').itervalues():
            for profile in handler.query(query):
                node = profile.toXML(_doc)
                tmp.write(node.toxml())
        tmp.write(u'</profiles>')
        tmp.flush()
        tmp.seek(0L)
        self.send_response(200)
        self.send_header('Content-type', 'text/xml;charset=utf-8')
        size = os.fstat(tmp.fileno())[stat.ST_SIZE]
        self.send_header('Content-length', str(size))
        self.end_headers()
        shutil.copyfileobj(tmp, self.wfile)
        self.log_request(200, size)
    

class WebGridServer(HTTPServer):
    '''Web grid HTTP server.  This server handles incoming HTTP requests and directs them to a
    WebGridRequestHandler.  It also contains the server's ID, and the sequences of profile and
    product handlers.
    '''
    def __init__(self, addr, serverID):
        '''Initialize by saving the server ID and creating empty sequences of profile
        and product handlers.
        '''
        HTTPServer.__init__(self, addr, WebGridRequestHandler)
        self.serverID = serverID
        self.__handlers = {}
        for kind in _validKinds:
            self.__handlers[kind] = {}
    
    def getHandlers(self, kind):
        '''Get the map of handlers for the given kind, which is either "product" or "profile".
        '''
        if kind not in _validKinds:
            raise ValueError('Invalid handler kind "%s"' % kind)
        return self.__handlers[kind]
    

def _main():
    '''Run the web grid server.
    '''
    import sys
    try:
        serverID = sys.argv[1]
    except IndexError:
        serverID = 'oodt'
    listenAddr = ('', 7576)
    httpd = WebGridServer(listenAddr, serverID)
    httpd.serve_forever()
    

if __name__ == '__main__':
    _main()



# encoding: utf-8
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE.txt file distributed with
# this work for additional information regarding copyright ownership.  The ASF
# licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.    

'''
Agile OODT Query Expressions.

Query expressions in OODT are based on the "DIS" style expressions originally
developed for the Planetary Data System.  They consist of
keyword/operator/literal-value triples, such as `targetName = Mars`, each
linked with logical operators (and, or, not) and grouped with parentheses.
For more information, see OODT_.

This module defines classes that model the aspects of a query.  In general,
use the `Query` class, passing a string containing your keyword expression as
the first constructor argument.  From there, you have a `Query` object you can
pass around to profile servers, product servers, and so forth.

.. _OODT: http://incubator.apache.org/projects/oodt.html
'''

__docformat__ = 'restructuredtext'

import oodterrors, shlex, xmlutils, xml.dom
from xmlutils import DocumentableField

class _QueryExpressionScanner(shlex.shlex):
    '''Extend the shlex scanner but for the DIS-style query expressions we expect.
    This means adding a dot to the characters that comprise a word so we can easily parse
    floating-point numbers.  Also, there's no comment character.
    '''
    def __init__(self, str):
        '''Create scanner.  `str` is the string to scan.
        '''
        shlex.shlex.__init__(self, str)
        self.commenters = ''
        self.wordchars = self.wordchars + '.-/:'
        

    def get_token(self):
        '''Get the next token.  We strip quotes from strings, attach negative signs
        to the numbers they're negating, and attach the = to <, >, and ! where needed.
        '''
        token = shlex.shlex.get_token(self)
        if token == self.eof or token == None:
            return None
        if token[0] in self.quotes:
            token = token[1:-1]
        elif token in ('<', '>', '!'):
            next = shlex.shlex.get_token(self)
            if next == self.eof or next == None:
                return None
            elif next == '=':
                token = token + next
            else:
                self.push_token(next)
        return token
    

class QueryException(oodterrors.OODTException):
    '''Exceptions related to query expression or query services
    '''
    pass
    

class ExpressionParseError(QueryException):
    '''Error in parsing a query expression.
    '''
    pass
    

class QueryElement(xmlutils.Documentable):
    '''An element of a query.
    '''
    def __init__(self, role='UNKNOWN', value='UNKNOWN', node=None):
        '''Create a QueryElement.  You can provide role and value settings, or provide an XML
        DOM node which will be parsed for role/value.
        '''
        self.role, self.value = role, value
        if node != None:
            self.parse(node)

    def getDocumentElementName(self):
        '''Give the XML tag name: `queryElement`.
        '''
        return 'queryElement'
    
    def getDocumentableFields(self):
        '''Get the attributes that go into XML.
        '''
        return (DocumentableField('role', u'tokenRole', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('value', u'tokenValue', DocumentableField.SINGLE_VALUE_KIND))
    
    def __repr__(self):
        return 'QueryElement(role="%s",value="%s")' % (self.role, self.value)
    

class QueryHeader(xmlutils.Documentable):
    '''Header of a query.  Captures metadata like data dictionary in use, etc.
    '''
    def __init__(self, id='UNKNOWN', title='UNKNOWN', desc='UNKNOWN', type='QUERY', status='ACTIVE', security='UNKNOWN',
        rev='2005-10-01 SCK v0.0.0 Under Development', dataDict='UNKNOWN', node=None):
        '''Initialize a QueryHeader.  Provide id, title, desc, type, status, security, rev,
        and dataDict settings.  Or, provide just an XML DOM node to be parsed.
        '''
        self.id, self.title, self.desc, self.type, self.status, self.security, self.rev, self.dataDict = \
            id, title, desc, type, status, security, rev, dataDict
        if node != None:
            self.parse(node)
    
    def getDocumentElementName(self):
        '''Give the XML tag name: `queryAttributes`.
        '''
        return 'queryAttributes'
    
    def getDocumentableFields(self):
        '''Get the attributes that go into XML.
        '''
        return (DocumentableField('id', u'queryId', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('title', u'queryTitle', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('desc', u'queryDesc', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('type', u'queryType', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('status', u'queryStatusId', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('security', u'querySecurityType', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('rev', u'queryRevisionNote', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('dataDict', u'queryDataDictId', DocumentableField.SINGLE_VALUE_KIND))
    
    def __repr__(self):
        return 'QueryHeader(id="%s",title="%s",desc="%s",type="%s",status="%s",security="%s",rev="%s",dataDict="%s")' % (
            self.id, self.title, self.desc, self.type, self.status, self.security, self.rev, self.dataDict
        )
    
    def __cmp__(self, other):
        return cmp((self.id, self.title, self.desc, self.type, self.status, self.security, self.rev, self.dataDict),
               (other.id, other.title, other.desc, other.type, other.status, other.security, other.rev, other.dataDict))
    

class QueryResult(object):
    '''Result of a query.
    '''
    def __init__(self, results=[], node=None):
        '''Results of a query are captured as a sequence of generic objects.
        '''
        self.results = results
        if node != None: self.parse(node)
    
    def parse(self, node):
        '''Initialize this object from the given XML DOM `node`.
        '''
        if 'queryResultSet' != node.nodeName:
            raise ValueError('Expected queryResultSet but got "%s"' % node.nodeName)
        for child in node.childNodes:
            if child.nodeType == xml.dom.Node.ELEMENT_NODE and 'resultElement' == child.nodeName:
                self.results.append(Result(node=child))
    
    def toXML(self, owner):
        '''Convert this object into XML owned by the given `owner` document.
        '''
        root = owner.createElement('queryResultSet')
        for result in self.results:
            root.appendChild(result.toXML(owner))
        return root
    
    def clear(self):
        '''Remove all results.
        '''
        self.results = []
    
    def __getitem__(self, i):
        return self.results[i]
    
    def __len__(self):
        return len(self.results)
    
    def __cmp__(self, other):
        return cmp(self.results, other.results)
    

_RELOPS = {
    'LT':      'LT', 
    'lt':      'LT', 
    '<':       'LT',
    'LE':      'LE',
    'le':      'LE',
    '<=':      'LE',
    'EQ':      'EQ', 
    'eq':      'EQ', 
    '=':       'EQ',
    'GE':      'GE',
    'ge':      'GE',
    '>=':      'GE',
    'GT':      'GT',
    'gt':      'GT',
    '>':       'GT',
    'NE':      'NE',
    'ne':      'NE',
    '!=':      'NE',
    'LIKE':    'LIKE',
    'like':    'LIKE',
    'NOTLIKE': 'NOTLIKE',
    'notlike': 'NOTLIKE',
    'notLike': 'NOTLIKE',
    'IS':      'IS',
    'is':      'is',
    'ISNOT':   'ISNOT',
    'isnot':   'isnot',
    'isNot':   'isnot'
}

_LOGOPS = {
    'AND': 'AND',
    'and': 'AND',
    '&':   'AND',
    'OR':  'OR',
    'or':  'OR',
    '|':   'OR',
    'NOT': 'NOT',
    'not': 'NOT',
    '!':   'NOT'
}

_PRECEDENCE = {
    'NOT': 2,
    'AND': 1,
    'OR': 0,
}

class Query(object):
    '''Query.  In old OODT, this was called XMLQuery, even though XML was tagential to it.
    Captures aspects of a query, including header, results, and so forth.  Most importantly, it
    captures the query expression, which contains the constraints on user's desiderata and
    range on what to return.
    
    As with other classes in this module, you can provide an XML DOM node to be parsed
    for the query's settings, or provide each of them individually.
    '''
    def __init__(self, keywordQuery=None, header=QueryHeader(), resultModeID='ATTRIBUTE', propType='BROADCAST',
        propLevels='N/A', maxResults=1, mimeAccept=[], parseQuery=True, node=None):
        '''Initialize a query.  Usually you provide just the `keywordQuery` argument
        which should be a keyword/query expression in the DIS-style; or you provide
        the `node` which is an XML DOM node describing a query.
        '''
        self.header, self.resultModeID, self.propType, self.propLevels, self.maxResults, self.mimeAccept = \
            header, resultModeID, propType, propLevels, maxResults, mimeAccept
        self.wheres, self.selects, self.froms, self.resultSet = [], [], [], QueryResult()
        if keywordQuery != None:
            self.keywordQuery = keywordQuery
            if parseQuery:
                self.wheres, self.selects = _parseQuery(keywordQuery)
        else:
            self.keywordQuery = ''
        if node != None:
            self.parse(node)
        
    def toXML(self, owner):
        '''Yield this query as an XML DOM.
        '''
        query = owner.createElement('query')
        query.appendChild(self.header.toXML(owner))
        xmlutils.add(query, u'queryResultModeId', self.resultModeID)
        xmlutils.add(query, u'queryPropogationType', self.propType)
        xmlutils.add(query, u'queryPropogationLevels', self.propLevels)
        for mimeType in self.mimeAccept:
            xmlutils.add(query, u'queryMimeAccept', mimeType)
        xmlutils.add(query, u'queryMaxResults', str(self.maxResults))
        xmlutils.add(query, u'queryKWQString', self.keywordQuery)
        selects = owner.createElement(u'querySelectSet')
        query.appendChild(selects)
        for select in self.selects:
            selects.appendChild(select.toXML(owner))
        fromElement = owner.createElement(u'queryFromSet')
        query.appendChild(fromElement)
        for i in self.froms:
            fromElement.appendChild(i.toXML(owner))
        wheres = owner.createElement(u'queryWhereSet')
        query.appendChild(wheres)
        for where in self.wheres:
            wheres.appendChild(where.toXML(owner))
        query.appendChild(self.resultSet.toXML(owner))
        return query
        
    def parse(self, node):
        '''Parse the XML DOM node as a query document.
        '''
        if 'query' != node.nodeName:
            raise ValueError('Expected query but got "%s"' % node.nodeName)
        self.mimeAccept, self.results = [], []
        for child in node.childNodes:
            if child.nodeType == xml.dom.Node.ELEMENT_NODE:
                if child.nodeName == u'queryAttributes':
                    self.header = QueryHeader(node=child)
                elif child.nodeName == u'resultModeID':
                    self.resultModeID = xmlutils.text(child)
                elif child.nodeName == u'queryPropogationType':
                    self.propType = xmlutils.text(child)
                elif child.nodeName == u'queryPropogationLevels':
                    self.propLevels = xmlutils.text(child)
                elif child.nodeName == u'queryMimeAccept':
                    self.mimeAccept.append(xmlutils.text(child))
                elif child.nodeName == u'queryMaxResults':
                    self.maxResults = int(xmlutils.text(child))
                elif child.nodeName == u'queryKWQString':
                    self.keywordQuery = xmlutils.text(child)
                elif child.nodeName == u'querySelectSet':
                    self.selects = _parseQueryElements(child)
                elif child.nodeName == u'queryFromSet':
                    self.froms = _parseQueryElements(child)
                elif child.nodeName == u'queryWhereSet':
                    self.wheres = _parseQueryElements(child)
                elif child.nodeName == u'queryResultSet':
                    self.resultSet = QueryResult(node=child)
    
    def __cmp__(self, other):
        header = cmp(self.header, other.header)
        if header < 0:
            return -1
        elif header == 0:
            resultModeID = cmp(self.resultModeID, other.resultModeID)
            if resultModeID < 0:
                return -1
            elif resultModeID == 0:
                propType = cmp(self.propType, other.propType)
                if propType < 0:
                    return -1
                elif propType == 0:
                    propLevels = cmp(self.propLevels, other.propLevels)
                    if propLevels < 0:
                        return -1
                    elif propLevels == 0:
                        maxResults = self.maxResults - other.maxResults
                        if maxResults < 0:
                            return -1
                        elif maxResults == 0:
                            mimeAccept = cmp(self.mimeAccept, other.mimeAccept)
                            if mimeAccept < 0:
                                return -1
                            elif mimeAccept == 0:
                                selects = cmp(self.selects, other.selects)
                                if selects < 0:
                                    return -1
                                elif selects == 0:
                                    froms = cmp(self.froms, other.froms)
                                    if froms < 0:
                                        return -1
                                    elif froms == 0:
                                        wheres = cmp(self.wheres, other.wheres)
                                        if wheres < 0:
                                            return -1
                                        elif wheres == 0:
                                            return cmp(self.resultSet, other.resultSet)
        return 1
    

def _parseQueryElements(node):
    '''The children of the given XML DOM node are a sequence of queryElements.  Parse them
    and return a list of QueryElement objects.
    '''
    a = []
    for child in node.childNodes:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            a.append(QueryElement(node=child))
    return a
    

def _parseQuery(s):
    '''Parse the query expression in `s`.
    '''
    if s is None:
        return [], []
    if len(s.strip()) == 0:
        return [], []
    if s.count('(') != s.count(')'):
        raise ExpressionParseError('Unbalanced parentheses')
    scanner = _QueryExpressionScanner(s)
    return _buildQuery(scanner)
    

def _buildQuery(scanner):
    '''Build the query stacks using the given `scanner`.
    '''
    operators, expression, selectors = [], [], []
    while True:
        token = scanner.get_token()
        if token is None: break
        if token in _LOGOPS:
            op = QueryElement('LOGOP', _LOGOPS[token])
            if len(operators) == 0:
                operators.append(op)
            else:
                while len(operators) > 0 and _PRECEDENCE[operators[-1].value] > _PRECEDENCE[op.value]:
                    expression.append(operators.pop())
                operators.append(op)
        elif token == '(':
            subExpr, subSelectors = _buildQuery(scanner)
            expression.extend(subExpr)
            selectors.extend(subSelectors)
        elif token == ')':
            break
        else:
            _addTerm(token, scanner, expression, operators, selectors)
    if len(operators) > 0 and len(expression) == 0:
        raise ExpressionParseError('Query contains only logical operators')
    operators.reverse()
    expression.extend(operators)
    return expression, selectors
    

def _addTerm(elemName, scanner, expression, operators, selectors):
    '''Add a term to the correct stack.
    '''
    relop = scanner.get_token()
    if relop is None:
        raise ExpressionParseError('Expected relational operator after element name "%s"' % elemName)
    if relop not in _RELOPS:
        raise ExpressionParseError('Unknown relational operator "%s"' % relop)
    literal = scanner.get_token()
    if literal is None:
        raise ExpressionParseError('Expected literal value for "%s %s" comparison' % (elemName, relop))
    if elemName == 'RETURN':
        selectors.append(QueryElement('elemName', literal))
        if len(operators) > 0:
            operators.pop()
        else:
            scanner.get_token()
    else:
        expression.append(QueryElement('elemName', elemName))
        expression.append(QueryElement('LITERAL', literal))
        expression.append(QueryElement('RELOP', _RELOPS[relop]))
    

# Sample code:
# if __name__ == '__main__':
#   import urllib, xml.dom.minidom
# 
#   impl = xml.dom.minidom.getDOMImplementation()
#   doc = impl.createDocument(None, None, None) # nsURI, qName, docType
# 
#   q = Query('track = Innocente')
#   node = q.toXML(doc)
#   doc.appendChild(node)
#   q = doc.toxml()
#   f = urllib.urlopen('http://localhost:8080/pds/prof', urllib.urlencode(dict(xmlq=q)), {}) # url, postdata, proxies (none)
#   print f.read()



# encoding: utf-8
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE.txt file distributed with
# this work for additional information regarding copyright ownership.  The ASF
# licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.    

'''Profiles.  A profile is a metadata description of a resource.

Profiles capture:

* Inception metadata.  This includes items such as a resouce's title,
  description, creators, contributors, publishers, language, and so on.  This
  set of metdata is based on the Dublin Core.  Class `ResourceAttributes`
  captures this metadata.

* Composition metadata.  This includes data elements that describe what
  the resource contains, such as ranges of data values (high and low
  temperature, latitude/longitude) or enumerated data values (zoning code,
  genus/species, etc.).  Subclasses of `ProfileElement` captures this
  metadata.

* Profile metadata.  This is metadata describing the profile itself, such
  as revision notes, ID number, etc.  Class `ProfileAttributes` captures this
  metadata.

Objects of these classes are unified into a single `Profile` object.
'''

__docformat__ = 'restructuredtext'

import xmlutils
import xml.dom
from xmlutils import DocumentableField

class ProfileAttributes(xmlutils.Documentable):
    '''Attributes of a profile.  These are attributes not related to the
    resource that a profile profiles, but rather the profile itself.  In
    most cases, simply constructing this object with no initializer
    arguments will suffice.
    '''
    def __init__(self, id='UNKNOWN', version='1.0.0', type='profile', statusID='active', securityType='unclassified',
        parentID='UNKNOWN', childIDs=None, regAuthority='UNKNOWN', revNotes=None, node=None):
        '''Initialize profile attributes.
        
        - `id` unique ID for the profile
        - `version` number
        - `type` should always be the string "profile"
        - `statusID` should always be the string "active"
        - `securityType` tells whether the profile is secret
        - `parentID` gives the unique ID for any "parent" profile
        - `childIDs` should be a sequnece of any "children" profiles
        - `regAuthority` is a unique ID identifying the registration
           authority responsible for the profile
        - `revNotes` is a sequence of comments detailing historical
           changes to the profile
        - `node` is an XML DOM node.  If the `node` argument is given,
           it's used to initialize the object.
        '''
        self.id = id
        self.version = version
        self.type = type
        self.statusID = statusID
        self.securityType = securityType
        self.parentID = parentID
        self.childIDs = childIDs
        self.regAuthority = regAuthority
        self.revNotes = revNotes
        if self.childIDs is None:
            self.childIDs = []
        if self.revNotes is None:
            self.revNotes = []
        if node is not None:
            self.parse(node)
    
    def getDocumentElementName(self):
        '''Get the XML tag for objects of this class: `profAttributes`.
        '''
        return 'profAttributes'
    
    def getDocumentableFields(self):
        '''Get the attributes that are put into XML.
        '''
        return (DocumentableField('id', u'profId', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('version', u'profVersion', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('type', u'profType', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('statusID', u'profStatusId', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('securityType', u'profSecurityType', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('parentID', u'profParentId', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('childIDs', u'profChildId', DocumentableField.MULTI_VALUED_KIND),
            DocumentableField('regAuthority', u'profRegAuthority', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('revNotes', u'profRevisionNote', DocumentableField.MULTI_VALUED_KIND))

    def __repr__(self):
        return ('ProfileAttributes(id=%s,version=%s,type=%s,statusID=%s,securityType=%s,parentID=%s,childIDs=%s,' \
            'regAuthority=%s,revNotes=%s)') % (self.id, self.version, self.type, self.statusID, self.securityType,
            self.parentID, self.childIDs, self.regAuthority, self.revNotes)
            
    def __cmp__(self, other):
        return cmp(self.id, other.id)
    
    def __hash__(self):
        return hash(self.id)
    

class ResourceAttributes(xmlutils.Documentable):
    '''Attributes of the resource.  Objects of this class collect data about the resource's
    inception and are based on Dublin Core.
    '''
    def __init__(self, identifier='UNKNOWN', title='UNKNOWN', formats=None, description='UNKNOWN', creators=None, subjects=None,
        publishers=None, contributors=None, dates=None, types=None, sources=None, languages=None, relations=None,
        coverages=None, rights=None, contexts=None, aggregation='UNKNOWN', resClass='UNKNOWN', locations=None, node=None):
        '''Initialize ResourceAttributes.
        
        The following arguments are required:
        
        - `identifier` is a URI uniquely identifying the resource
        - `title` names the resource
        - `description` gives a summary or abstract of it
        - `aggregation` tells the gross structure of the resource and
           should be one of the following values:
            - data.granule
            - data.dataSet
            - data.dataSetCollection
        - `resClass` gives the kind of the resource (what to expect
           when connecting to one of its locations)
        
        All of the others should be initialized with sequences as the
        resource attributes follows the Dublin Core recommendation on
        multiplicity.
        
        If `node` is given, it's treated as an XML DOM node and is used to
        initialize the resource attributes.
        '''
        self.identifier = identifier
        self.title = title
        self.formats = formats
        self.description = description
        self.creators = creators
        self.subjects = subjects
        self.publishers = publishers
        self.contributors = contributors
        self.dates = dates
        self.types = types
        self.sources = sources
        self.languages = languages
        self.relations = relations
        self.coverages = coverages
        self.rights = rights
        self.contexts = contexts
        self.aggregation = aggregation
        self.resClass = resClass
        self.locations = locations
        for attr in ('formats', 'creators', 'subjects', 'publishers', 'contributors', 'dates', 'sources', 'languages',
            'relations', 'coverages', 'rights', 'contexts', 'locations', 'types'):
            if getattr(self, attr) is None:
                setattr(self, attr, [])
        if node is not None:
            self.parse(node)
    
    def getDocumentElementName(self):
        '''Give the XML tag name: `resAttributes`.
        '''
        return 'resAttributes'
        
    def __hash__(self):
        return hash(self.identifier)
    
    def __cmp__(self, other):
        return cmp(self.identifier, other.identifier)
    
    def getDocumentableFields(self):
        '''Get the attributes that go into XML.
        '''
        return (DocumentableField('identifier', u'Identifier', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('title', u'Title', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('formats', u'Format', DocumentableField.MULTI_VALUED_KIND),
            DocumentableField('description', u'Description', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('creators', u'Creator', DocumentableField.MULTI_VALUED_KIND),
            DocumentableField('subjects', u'Subject', DocumentableField.MULTI_VALUED_KIND),
            DocumentableField('publishers', u'Publisher', DocumentableField.MULTI_VALUED_KIND),
            DocumentableField('contributors', u'Contributor', DocumentableField.MULTI_VALUED_KIND),
            DocumentableField('dates', u'Date', DocumentableField.MULTI_VALUED_KIND),
            DocumentableField('types', u'Type', DocumentableField.MULTI_VALUED_KIND),
            DocumentableField('sources', u'Source', DocumentableField.MULTI_VALUED_KIND),
            DocumentableField('languages', u'Language', DocumentableField.MULTI_VALUED_KIND),
            DocumentableField('relations', u'Relation', DocumentableField.MULTI_VALUED_KIND),
            DocumentableField('coverages', u'Coverage', DocumentableField.MULTI_VALUED_KIND),
            DocumentableField('rights', u'Rights', DocumentableField.MULTI_VALUED_KIND),
            DocumentableField('contexts', u'resContext', DocumentableField.MULTI_VALUED_KIND),
            DocumentableField('aggregation', u'resAggregation', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('resClass', u'resClass', DocumentableField.SINGLE_VALUE_KIND),
            DocumentableField('locations', u'resLocation', DocumentableField.MULTI_VALUED_KIND))
    

class Profile(object):
    '''A profile "profiles" a resource by describing it with metadata.
    '''
    def __init__(self, profAttr=None, resAttr=None, profElements=None, node=None):
        '''Initialize a profile.  The profElements should be a dicitonary that maps
        from element name to instance of a `ProfileElement`.
        '''
        self.profAttr, self.resAttr, self.profElements = profAttr, resAttr, profElements
        if self.profAttr is None:
            self.profAttr = ProfileAttributes()
        if self.resAttr is None:
            self.resAttr = ResourceAttributes()
        if self.profElements is None:
            self.profElements = {}
        if node is not None:
            self.parse(node)
    
    def parse(self, node):
        '''Initialize this object from the given XML DOM `node`.
        '''
        if node.nodeName != 'profile':
            raise ValueError('Expected profile element but got "%s"' % node.nodeName)
        for child in filter(lambda node: node.nodeType == xml.dom.Node.ELEMENT_NODE, node.childNodes):
            name = child.nodeName
            if name == u'profAttributes':
                self.profAttr = ProfileAttributes()
                self.profAttr.parse(child)
            elif name == u'resAttributes':
                self.resAttr = ResourceAttributes()
                self.resAttr.parse(child)
            elif name == u'profElement':
                elem = _parseProfileElement(child)
                self.profElements[elem.name] = elem
                
    def toXML(self, owner):
        '''Convert this object into XML owned by the given `owner` document.
        '''
        root = owner.createElement(u'profile')
        root.appendChild(self.profAttr.toXML(owner))
        root.appendChild(self.resAttr.toXML(owner))
        for elem in self.profElements.itervalues():
            root.appendChild(elem.toXML(owner))
        return root
    
    def __cmp__(self, other):
        profAttr = cmp(self.profAttr, other.profAttr)
        if profAttr < 0:
            return -1
        elif profAttr == 0:
            resAttr = cmp(self.resAttr, other.resAttr)
            if resAttr < 0:
                return -1
            elif resAttr == 0:
                return cmp(self.profElements, other.profElements)
        return 1
    
    def __hash__(self):
        return hash(self.profAttr) ^ hash(self.resAttr)
    
    def __repr__(self):
        return 'Profile(profAttr=%s,resAttr=%s,profElements=%s)' % (self.profAttr, self.resAttr, self.profElements)
    

class ProfileElement(object):
    '''Abstract profile element.
    '''
    def __init__(self, name, description, type, units, synonyms, comment):
        '''Initialize profile element.
        '''
        self.name = name
        self.description = description
        self.type = type
        self.units = units
        self.synonyms = synonyms
        self.comment = comment
    
    def __repr__(self):
        return 'ProfileElement(name=%r,description=%r,type=%r,units=%r,synonyms=%r,comment=%r)' % (self.name,
            self.description, self.type, self.units, self.synonyms, self.comment)
    
    def __cmp__(self, other):
        rc = cmp(self.name, other.name)
        if rc < 0:
            return -1
        elif rc == 0:
            rc = cmp(self.description, other.description)
            if rc < 0:
                return -1
            elif rc == 0:
                rc = cmp(self.type, other.type)
                if rc < 0:
                    return -1
                elif rc == 0:
                    rc = cmp(self.units, other.units)
                    if rc < 0:
                        return -1
                    elif rc == 0:
                        rc = cmp(self.synonyms, other.synonyms)
                        if rc < 0:
                            return -1
                        elif rc == 0:
                            return cmp(self.comment, other.comment)
        return 1
    
    def __hash__(self):
        return reduce(lambda x, y: hash(x) ^ hash(y), [getattr(self, attr) for attr in ('name', 'description', 'type',
            'units', 'synonyms', 'comment')])
    
    def isEnumerated(self):
        '''Is this an enumerated element?  Enumerated elements include those with
        discrete values as well unspecified elements.
        '''
        return False;
    
    def getValues(self):
        '''Get the discrete values of this element, which may be empty.
        '''
        return []
    
    def getMinValue(self):
        '''Get the minimum value of this element, which may be zero.
        '''
        return 0.0
    
    def getMaxValue(self):
        '''Get the maximum value of this element, which may be zero.
        '''
        return 0.0
    
    def toXML(self, owner):
        '''Convert this object into XML owned by the given `owner` document.
        '''
        root = owner.createElement('profElement')
        xmlutils.add(root, u'elemId', self.name)
        xmlutils.add(root, u'elemName', self.name)
        xmlutils.add(root, u'elemDesc', self.description)
        xmlutils.add(root, u'elemType', self.type)
        xmlutils.add(root, u'elemUnit', self.units)
        if self.isEnumerated():
            flag = 'T'
        else:
            flag = 'F'
        xmlutils.add(root, u'elemEnumFlag', flag)
        for value in self.getValues():
            elem = owner.createElement('elemValue')
            root.appendChild(elem)
            elem.appendChild(owner.createCDATASection(value))
        if not self.isEnumerated():
            xmlutils.add(root, u'elemMinValue', str(self.getMinValue()))
            xmlutils.add(root, u'elemMaxValue', str(self.getMaxValue()))
        xmlutils.add(root, u'elemSynonym', self.synonyms)
        xmlutils.add(root, u'elemComment', self.comment)
        return root
    

class UnspecifiedProfileElement(ProfileElement):
    '''An unspecified profile element merely documents the existence of a element within a
    dataset but says nothing about its actual values.
    '''
    def __init__(self, name, description, type, units, synonyms, comment):
        '''Initialize an "unspecified" profile element.
        '''
        super(UnspecifiedProfileElement, self).__init__(name, description, type, units, synonyms, comment)
    
    def isEnumerated(self):
        '''An unspecified profile element is indeed enumerated.  It just has no enumerated values.
        '''
        return True
    

class EnumeratedProfileElement(ProfileElement):
    '''An enumerated profile element describes set of discrete values.
    '''
    def __init__(self, name, description, type, units, synonyms, comment, values):
        '''Initialize an enumerated profile element.
        '''
        super(EnumeratedProfileElement, self).__init__(name, description, type, units, synonyms, comment)
        self.values = values
        
    def isEnumerated(self):
        '''An enumerated profile element is indeed enumerated.
        '''
        return True
    
    def getValues(self):
        '''Return the sequence of values.
        '''
        return self.values
        
    def __cmp__(self, other):
        rc = super(EnumeratedProfileElement, self).__cmp__(other)
        if rc < 0:
            return -1
        elif rc == 0:
            return cmp(self.values, other.values)
        return 1
    

class RangedProfileElement(ProfileElement):
    '''A ranged profile element describes a value between two numeric ranges.
    '''
    def __init__(self, name, description, type, units, synonyms, comment, minValue, maxValue):
        '''Initialize a ranged profile element.
        '''
        super(RangedProfileElement, self).__init__(name, description, type, units, synonyms, comment)
        self.minValue = minValue
        self.maxValue = maxValue
    
    def getMinValue(self):
        '''Get the minimum value.
        '''
        return self.minValue
    
    def getMaxValue(self):
        '''Get the maximum value.
        '''
        return self.maxValue
    
    def __repr__(self):
        return 'RangedProfileElement(%r,minValue=%r,maxValue=%r)' % (super(RangedProfileElement, self).__repr__(),
            self.minValue, self.maxValue)
    
    def __cmp__(self, other):
        rc = super(RangedProfileElement, self).__cmp__(other)
        if rc < 0:
            return -1
        elif rc == 0:
            rc = self.minValue - other.minValue
            if rc < 0:
                return -1
            elif rc == 0:
                return self.maxValue - other.maxValue
        return 1
    

def _parseProfileElement(node):
    '''Construct an appropriate profile element from the given DOM node.
    '''
    if node.nodeName != u'profElement':
        raise ValueError('Expected profElement element but got "%s"' % node.nodeName)
    settings = dict(elemId=u'UNKNOWN', elemDesc=u'UNKNOWN', elemType=u'UNKNOWN', elemUnit=u'UNKNOWN', elemEnumFlag=u'F',
        elemComment=u'UNKNOWN')
    values, syns = [], []
    for child in filter(lambda node: node.nodeType == xml.dom.Node.ELEMENT_NODE, node.childNodes):
        text = xmlutils.text(child)
        if child.nodeName == u'elemValue':
            values.append(text)
        elif child.nodeName == u'elemSynonym':
            syns.append(text)
        else:
            settings[str(child.nodeName)] = text
    if 'elemName' not in settings:
        raise ValueError('profElement requires elemName but none specified')
    if 'elemEnumFlag' not in settings:
        raise ValueError('profElement requires elemEnumFlag but none specified')
        
    # Normally I'd treat only those XML elements where elemEnumFlag as T as possibly producing
    # unspecified or enumerated, and F producing *only* ranged elements.  But PDS profile
    # servers are producing profile elements with F for elemEnumFlag and yet NO elemMinValue
    # and elemMaxValue.  If they're using the Java profile code, I'd call that a bug in that
    # code.  If they're not, then I'd say they're producing profiles incorrectly.
    if settings['elemEnumFlag'] == 'T':
        if len(values) == 0:
            return UnspecifiedProfileElement(settings['elemName'], settings['elemDesc'], settings['elemType'],
                settings['elemUnit'], syns, settings['elemComment'])
        else:
            return EnumeratedProfileElement(settings['elemName'], settings['elemDesc'], settings['elemType'],
                settings['elemUnit'], syns, settings['elemComment'], values)
    else:
        if 'elemMinValue' not in settings or 'elemMaxValue' not in settings:
            return UnspecifiedProfileElement(settings['elemName'], settings['elemDesc'], settings['elemType'],
                settings['elemUnit'], syns, settings['elemComment'])
        else:
            return RangedProfileElement(settings['elemName'], settings['elemDesc'], settings['elemType'],
            settings['elemUnit'], syns, settings['elemComment'], float(settings['elemMinValue']),
            float(settings['elemMaxValue']))
    

# Some sample code:
# if __name__ == '__main__':
#   import urllib2, xml.dom.minidom
#   x = urllib2.urlopen('http://starbrite.jpl.nasa.gov/q?object=urn%3Aeda%3Armi%3AJPL.PDS.MasterProd&type=profile&keywordQuery=TARGET_NAME+%3D+MARS')
#   d = xml.dom.minidom.parse(x)
#   x.close()
#   profiles = []
#   for i in d.documentElement.getElementsByTagName(u'profile'):
#       profiles.append(Profile(node=i))
#       
#   doc = xml.dom.minidom.getDOMImplementation().createDocument(None, None, None)
#   print '<?xml version="1.0" encoding="UTF-8"?>'
#   print '<!DOCTYPE profiles PUBLIC "-//JPL//DTD Profile 1.1//EN"'
#   print '  "http://oodt.jpl.nasa.gov/grid-profile/dtd/prof.dtd">'
#   print '<profiles>'
#   for profile in profiles:
#       print profile.toXML(doc).toxml()
#   print '</profiles>'
#   



# encoding: utf-8
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE.txt file distributed with
# this work for additional information regarding copyright ownership.  The ASF
# licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.    

'''Object Oriented Data Technology XML utilities. These are some simple utilities
that make working with rather obtuse DOM API less painful.
'''

__docformat__ = 'restructuredtext'

import string
import xml.dom
    
def text(node):
    '''Return the text under the given node.  Text and CDATA nodes simply return
    their content.  Otherwise, text is gathered in a depth-first left-to-right
    traversal of the node, building up text as we go along.
    
    >>> import xml.dom
    >>> domImpl = xml.dom.getDOMImplementation()
    >>> doc = domImpl.createDocument(None, None, None)
    >>> root = doc.createElement('root')
    >>> root.appendChild(doc.createTextNode('alpha')) # doctest: +ELLIPSIS
    <DOM Text node...>
    >>> text(root)
    'alpha'
    >>> child = root.appendChild(doc.createElement('child'))
    >>> text(child.appendChild(doc.createCDATASection('beta')))
    'beta'
    >>> text(root)
    'alphabeta'
    '''
    strings = []
    _text0(node, strings)
    return string.join(strings, '') # no separator
    

def _text0(node, strings):
    '''Add the text under the given node to the list of strings.
    '''
    if node.nodeType == xml.dom.Node.CDATA_SECTION_NODE:
        strings.append(node.nodeValue)
    elif node.nodeType == xml.dom.Node.TEXT_NODE:
        strings.append(node.nodeValue)
    for child in node.childNodes:
        _text0(child, strings)
    

def add(node, elementName, value=None):
    '''Add an element under the given node.  Optionally, the value text is added to the element.
    Returns the original node, handy for chaining calls together.  If the value is a sequence,
    then the element is added multiple times for each item in the sequence, using the item
    as the text.

    >>> from xml.dom.minidom import getDOMImplementation
    >>> doc = getDOMImplementation().createDocument(None, None, None)
    >>> root = doc.createElement('root')
    >>> root.toxml()
    '<root/>'
    >>> add(root, 'empty').toxml()
    '<root><empty/></root>'
    >>> add(root, 'full', 'of text').toxml()
    '<root><empty/><full>of text</full></root>'
    >>> add(root, 'item', [str(i) for i in range(1, 4)]).toxml()
    '<root><empty/><full>of text</full><item>1</item><item>2</item><item>3</item></root>'
    '''
    owner = node.ownerDocument
    if isinstance(value, basestring):
        elem = owner.createElement(elementName)
        node.appendChild(elem)
        elem.appendChild(owner.createTextNode(value))
    elif value is not None:
        for i in value:
            elem = owner.createElement(elementName)
            node.appendChild(elem)
            elem.appendChild(owner.createTextNode(i))
    else:
        node.appendChild(owner.createElement(elementName))
    return node
    

class DocumentableField(object):
    '''A documentable field is an attribute of an object that we can serialize into XML.
    '''
    SINGLE_VALUE_KIND = 1
    MULTI_VALUED_KIND = 2
    DOCUMENTABLE_KIND = 3

    def __init__(self, attrName, elemName, kind):
        '''Initialize a documentable field with attrName (name of attribute in object),
        elemName (name to use for XML element) and kind, which is one of the DOCUMENTABLE_*
        constants.
        '''
        self.attrName = attrName
        self.elemName = elemName
        if kind not in (DocumentableField.SINGLE_VALUE_KIND, DocumentableField.MULTI_VALUED_KIND,
            DocumentableField.DOCUMENTABLE_KIND):
            raise ValueError('Invalid kind %d' % kind)
        self.kind = kind
    
    def __cmp__(self, other):
        attrName = cmp(self.attrName, other.attrName)
        if attrName < 0:
            return -1
        elif attrName == 0:
            elemName = cmp(self.elemName, other.elemName)
            if elemName < 0:
                return -1
            elif elemName == 0:
                return cmp(self.kind, other.kind)
        return 1
    
    def __hash__(self):
        return hash(self.attrName) ^ hash(self.elemName) ^ self.kind
    

class Documentable(object):
    '''An object that can be documented into XML and parsed from an XML document.
    '''
    def getDocumentElementName(self):
        '''Get the XML tag name.  Subclasses must override this
        otherwise they get the tag name `UNKNOWN`.
        '''
        return 'UNKNOWN'
        
    def getDocumentableFields(self):
        '''Get the sequence of documentable attributes.  Subclasses should override this
        with a sequence of `DocumentableField` objects.  By default, we return an empty
        sequence.
        '''
        return []
    
    def toXML(self, owner):
        '''Convert this object into XML owned by the given `owner` document.
        '''
        root = owner.createElement(self.getDocumentElementName())
        for df in self.getDocumentableFields():
            if df.kind == DocumentableField.SINGLE_VALUE_KIND:
                add(root, df.elemName, str(getattr(self, df.attrName)))
            elif df.kind == DocumentableField.MULTI_VALUED_KIND:
                add(root, df.elemName, [str(value) for value in getattr(self, df.attrName)])
            else:
                root.appendChild(getattr(self, df.attrName).toXML(owner))
        return root
    
    def computeValueFromDocument(self, attrName, text):
        '''Compute a value for the attribute named `attrName` from the
        XML text representation `text`.  Subclasses may wish to
        override this in order to provide custom typing for certain
        attributes, such as returning an integer or datetime value by
        parsing the text.  By default, the `text` is returned
        unmodified as a string.
        '''
        return text
    
    def parse(self, node):
        '''Initialize this object from the given XML DOM `node`.
        '''
        if node.nodeName != self.getDocumentElementName():
            raise ValueError('Expected %s element but got %s' % (self.getDocumentElementName(), node.nodeName))
        fieldMap = dict(zip([i.elemName for i in self.getDocumentableFields()], self.getDocumentableFields()))
        for child in filter(lambda n: n.nodeType == xml.dom.Node.ELEMENT_NODE, node.childNodes):
            name = child.nodeName
            if name in fieldMap:
                field = fieldMap[name]
                if field.kind == DocumentableField.SINGLE_VALUE_KIND:
                    setattr(self, field.attrName, self.computeValueFromDocument(field.attrName, text(child)))
                elif field.kind == DocumentableField.MULTI_VALUED_KIND:
                    if not hasattr(self, field.attrName):
                        setattr(self, field.attrName, [])
                    getattr(self, field.attrName).append(self.computeValueFromDocument(field.attrName,
                        text(child)))
                else:
                    getattr(self, field.attrName).parse(child)
    
    def __cmp__(self, other):
        '''The documentable fields provide a bonus: we can use them to do comparisons.
        '''
        for field in self.getDocumentableFields():
            attrName = field.attrName
            mine = getattr(self, attrName)
            others = getattr(other, attrName)
            rc = cmp(mine, others)
            if rc < 0:
                return -1
            elif rc > 0:
                return 1
        return 0
    
    def __hash__(self):
        '''The documentable fields provide another bonus: we can use them
        to do hashing.
        '''
        return reduce(lambda x, y: hash(x) ^ hash(y), [getattr(self, i.attrName) for i in self.getDocumentableFields()],
            0x55555555)
    


# encoding: utf-8
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE.txt file distributed with
# this work for additional information regarding copyright ownership.  The ASF
# licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.    

import os.path
from distribute_setup import use_setuptools
use_setuptools()
from setuptools import find_packages, setup


# Package data
# ------------

_name         = 'oodt'
_version      = '0.3'
_description  = 'Agile OODT'
_url          = 'http://oodt.apache.org/'
_downloadURL  = 'http://pypi.python.org/pypi/oodt/'
_author       = 'Sean Kelly'
_authorEmail  = 'kelly@apache.org'
_license      = 'Apache 2.0'
_namespaces   = []
_testSuite    = 'oodt.tests.test_suite'
_zipSafe      = True
_keywords     = 'data grid discovery query optimization object middleware archive catalog index'
_requirements = []
_entryPoints  = {
    'console_scripts': ['webgrid = oodt.webgrid:_main'],
}
_classifiers  = [
    'Environment :: Console',
    'Environment :: No Input/Output (Daemon)',
    'Intended Audience :: Developers',
    'Intended Audience :: Information Technology',
    'Intended Audience :: Science/Research',
    'Topic :: Database :: Front-Ends',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
    'Topic :: Internet :: Z39.50',
    'Topic :: Scientific/Engineering',
    'Development Status :: 5 - Production/Stable',
    'Environment :: Web Environment',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Internet :: WWW/HTTP',
    'Topic :: Software Development :: Libraries :: Python Modules',
]


# Setup Metadata
# --------------

def _read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

_header = '*' * len(_name) + '\n' + _name + '\n' + '*' * len(_name)
_longDescription = '\n\n'.join([
    _header,
    _read('README.txt'),
    _read('docs', 'INSTALL.txt'),
    _read('docs', 'HISTORY.txt')
])
open('doc.txt', 'w').write(_longDescription)

setup(
    author=_author,
    author_email=_authorEmail,
    classifiers=_classifiers,
    description=_description,
    download_url=_downloadURL,
    entry_points=_entryPoints,
    include_package_data=True,
    install_requires=_requirements,
    keywords=_keywords,
    license=_license,
    long_description=_longDescription,
    name=_name,
    namespace_packages=_namespaces,
    packages=find_packages(),
    test_suite=_testSuite,
    url=_url,
    version=_version,
    zip_safe=_zipSafe,
)



# encoding: utf-8
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE.txt file distributed with
# this work for additional information regarding copyright ownership.  The ASF
# licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.    

'''Unit tests for the `xmlutils` module.
'''

__docformat__ = 'restructuredtext'

import unittest, xml.dom, oodt.xmlutils
from oodt.xmlutils import DocumentableField

class XMLTest(unittest.TestCase):
    '''Unit test the XML utilities.
    '''
    def testText(self):
        '''Test getting text from nodes.
        '''
        domImpl = xml.dom.getDOMImplementation()
        doc = domImpl.createDocument(None, None, None) # ns URI, qual name, doctype
        root = doc.createElement('root')
        root.appendChild(doc.createTextNode('Hello'))
        child = doc.createElement('child')
        root.appendChild(child)
        root.appendChild(doc.createCDATASection('world'))
        child.appendChild(doc.createTextNode('cruel'))
        empty = doc.createElement('empty')
        root.appendChild(empty)

        self.assertEquals('cruel', oodt.xmlutils.text(child))
        self.assertEquals(0, len(oodt.xmlutils.text(empty)))
        self.assertEquals('Hellocruelworld', oodt.xmlutils.text(root))
    
    def testAdds(self):
        '''Test adding things to nodes.
        '''
        domImpl = xml.dom.getDOMImplementation()
        doc = domImpl.createDocument(None, None, None) # ns URI, qual name, doctype
        root = doc.createElement('root')

        oodt.xmlutils.add(root, 'first')
        self.assertEquals(0, len(oodt.xmlutils.text(root)))
        oodt.xmlutils.add(root, 'second', 'with text')
        self.assertEquals('with text', oodt.xmlutils.text(root))
        
        oodt.xmlutils.add(root, 'number', [str(i) for i in range(1, 4)])
        self.assertEquals('with text123', oodt.xmlutils.text(root))
    

class DocumentableTest(unittest.TestCase):
    '''Unit test the `Documentable` and `DocumentableField` classes.
    '''
    def testIt(self):
        '''Test the `Documentable` and `DocumentableField` classes.
        '''
        class X(oodt.xmlutils.Documentable):
            def __init__(self, mission='UNK', target='UNK', measurements=[], node=None):
                self.mission, self.target, self.measurements = mission, target, measurements
                if node is not None:
                    self.parse(node)
            def computeValueFromDocument(self, attrName, text):
                if attrName == 'measurements':
                    return int(text)
                else:
                    return text
            def getDocumentElementName(self):
                return 'X'
            def getDocumentableFields(self):
                return (DocumentableField('mission', u'MISSION', DocumentableField.SINGLE_VALUE_KIND),
                DocumentableField('target', u'TARGET', DocumentableField.SINGLE_VALUE_KIND),
                DocumentableField('measurements', u'MZR', DocumentableField.MULTI_VALUED_KIND))
            def __eq__(self, other):
                return (self.mission == other.mission and self.target == other.target
                    and cmp(self.measurements, other.measurements) == 0)
            def __str__(self):
                return '%s,%s:%s' % (self.mission, self.target, self.measurements)
        a = X('Explorer', 'Moon', [1, 2, 3])
        doc = xml.dom.getDOMImplementation().createDocument(None, None, None)
        node = a.toXML(doc)
        b = X(node=node)
        self.assertEquals(a, b)
    

def test_suite():
    '''Create the suite of tests.
    '''
    import doctest
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(XMLTest))
    suite.addTest(unittest.makeSuite(DocumentableTest))
    suite.addTest(doctest.DocTestSuite(oodt.xmlutils))
    return suite
    

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
